# More Information: https://cloud.google.com/cloud-build/docs/api
from dataclasses import dataclass, field
from typing import List, Dict, Any, Union
from urllib.parse import urlparse

from google.api_core.exceptions import AlreadyExists
from google.cloud.devtools import cloudbuild_v1

from gcp_pilot import exceptions
from gcp_pilot.base import GoogleCloudPilotAPI

TriggerType = cloudbuild_v1.BuildTrigger
AnyEventType = Union[cloudbuild_v1.GitHubEventsConfig, cloudbuild_v1.RepoSource]


class GoogleCloudBuild(GoogleCloudPilotAPI):
    _client_class = cloudbuild_v1.CloudBuildClient

    def make_build_step(
            self,
            name: str,
            identifier: str = None,
            args: list = None,
            env=None,
            entrypoint: str = None,
    ) -> cloudbuild_v1.BuildStep:
        return cloudbuild_v1.BuildStep(
            id=identifier,
            name=name,
            args=args,
            env=env,
            entrypoint=entrypoint,
        )

    def make_source_repo_event(
            self,
            repo_name: str,
            branch_name: str = None,
            tag_name: str = None,
            project_id: str = None,
    ) -> cloudbuild_v1.RepoSource:
        if not branch_name and not tag_name:
            branch_name = 'master'

        params = {}
        if branch_name:
            params['branch_name'] = branch_name
        if tag_name:
            params['tag_name'] = tag_name
        return cloudbuild_v1.RepoSource(
            project_id=project_id or self.project_id,
            repo_name=repo_name,
            **params,
        )

    def make_github_event(
            self,
            url,
            branch_name: str = None,
            tag_name: str = None,
    ) -> cloudbuild_v1.GitHubEventsConfig:
        if not branch_name and not tag_name:
            branch_name = 'master'

        params = {}
        if branch_name:
            params['branch'] = branch_name
        if tag_name:
            params['tag'] = tag_name

        path = urlparse(url).path
        owner, name = path.split('/')[1:]
        return cloudbuild_v1.GitHubEventsConfig(
            owner=owner,
            name=name,
            push=cloudbuild_v1.PushFilter(
                **params,
            ),
        )

    def _make_trigger(
            self,
            name: str,
            description: str,
            steps: List[cloudbuild_v1.BuildStep],
            event: AnyEventType,
            tags: List[str],
            images: List[str] = None,
            substitutions: Dict[str, str] = None,
    ) -> cloudbuild_v1.BuildTrigger:

        def _get_event_param():
            valid_events = {
                'trigger_template': cloudbuild_v1.RepoSource,
                'github': cloudbuild_v1.GitHubEventsConfig,
            }
            for key, klass in valid_events.items():
                if isinstance(event, klass):
                    return key
            raise exceptions.ValidationError(f"Unsupported event type {event.__class__.__name__,}")

        params = {
            _get_event_param(): event
        }

        return cloudbuild_v1.BuildTrigger(
            name=name,
            description=description,
            tags=tags,
            build=cloudbuild_v1.Build(
                steps=steps,
                images=images or [],
            ),
            substitutions=substitutions,
            **params,
        )

    async def get_trigger(self, trigger_id: str, project_id: str = None) -> TriggerType:
        response = self.client.get_build_trigger(
            trigger_id=trigger_id,
            project_id=project_id or self.project_id,
        )
        return response

    async def delete_trigger(self, trigger_id: str, project_id: str = None):
        response = self.client.delete_build_trigger(
            trigger_id=trigger_id,
            project_id=project_id or self.project_id,
        )
        return response

    async def create_trigger(
            self,
            name: str,
            description: str,
            event: AnyEventType,
            steps: List[cloudbuild_v1.BuildStep],
            tags: List[str] = None,
            project_id: str = None,
            images: List[str] = None,
            substitutions: Dict[str, str] = None,
    ) -> TriggerType:
        trigger = self._make_trigger(
            name=name,
            description=description,
            event=event,
            steps=steps,
            tags=tags,
            images=images,
            substitutions=substitutions,
        )

        response = self.client.create_build_trigger(
            trigger=trigger,
            project_id=project_id or self.project_id,
        )
        return response

    async def update_trigger(
            self,
            name: str,
            description: str,
            event: AnyEventType,
            steps: List[cloudbuild_v1.BuildStep],
            tags: List[str] = None,
            images: List[str] = None,
            substitutions: Dict[str, str] = None,
            project_id: str = None,
    ) -> TriggerType:
        trigger = self._make_trigger(
            name=name,
            description=description,
            event=event,
            steps=steps,
            tags=tags,
            images=images,
            substitutions=substitutions,
        )

        response = self.client.update_build_trigger(
            trigger_id=name,
            trigger=trigger,
            project_id=project_id or self.project_id,
        )
        return response

    async def create_or_update_trigger(
            self,
            name: str,
            description: str,
            event: AnyEventType,
            steps: List[cloudbuild_v1.BuildStep],
            tags: List[str] = None,
            project_id: str = None,
            images: List[str] = None,
            substitutions: Dict[str, str] = None,
    ) -> TriggerType:
        create_args = dict(
            name=name,
            description=description,
            event=event,
            steps=steps,
            tags=tags,
            project_id=project_id,
            images=images,
            substitutions=substitutions,
        )

        try:
            return await self.create_trigger(**create_args)
        except AlreadyExists:
            return await self.update_trigger(**create_args)

    def get_builds(self, trigger_id: str = None, project_id: str = None) -> cloudbuild_v1.Build:
        # https://cloud.google.com/cloud-build/docs/view-build-results#filtering_build_results_using_queries
        filters = []
        if trigger_id:
            filters.append(f'trigger_id="{trigger_id}"')

        all_builds = self.client.list_builds(
            filter=' AND '.join(filters),
            project_id=project_id or self.project_id,
        )
        for build in all_builds:
            yield build


@dataclass
class _SubstitutionVariable:
    key: str
    value: Any

    @property
    def as_kv(self) -> str:
        return '%s=%s' % (self.key, str(self))

    @property
    def custom_key(self) -> str:
        # Custom substitution variables must be prefixed with an underscore
        return f'_{self.key}'

    def __str__(self) -> str:
        # When used in a template, $_NAME should work, but it requires to be isolated by blank spaces
        # Thus, we use ${_NAME} by default, because it allows merging with other text.
        return '${%s}' % self.custom_key


@dataclass
class SubstitutionHelper:
    _variables: Dict[str, _SubstitutionVariable] = field(default_factory=dict)

    def add(self, **kwargs):
        for k, v in kwargs.items():
            variable = _SubstitutionVariable(key=k.upper(), value=v)
            self._variables[variable.key] = variable

    @property
    def as_dict(self) -> Dict[str, str]:
        return {
            variable.custom_key: str(variable.value)  # all values must be string or bytes
            for variable in self._variables.values()
        }

    def __getattr__(self, item: str):
        # Useful tool to ease the variable access (eg. substitution.MY_VAR_NAME)
        return self._variables[item.upper()]



class UnsupportedFormatException(Exception):
    pass


class ValidationError(Exception):
    pass


class UnboundException(Exception):
    pass


class NotFoundError(Exception):
    pass


class BigQueryJobError(Exception):
    def __init__(self, job):
        message = ' | '.join([error['message'] for error in job.errors])
        super().__init__(message)


class AlreadyExistsError(Exception):
    pass



from typing import Iterator, Dict, Any

from googleapiclient.errors import HttpError

from gcp_pilot import exceptions
from gcp_pilot.base import GoogleCloudPilotAPI

UserType = GroupType = MemberType = Dict[str, Any]


class Directory(GoogleCloudPilotAPI):
    _scopes = [
        'https://www.googleapis.com/auth/admin.directory.group',
        'https://www.googleapis.com/auth/admin.directory.user',
    ]

    def __init__(self, email: str, **kwargs):
        self.email = email
        super().__init__(
            serviceName='admin',
            version='directory_v1',
            cache_discovery=False,
            **kwargs,
        )

    def get_users(self, domain: str = None) -> Iterator[UserType]:
        yield from self._paginate(
            method=self.client.users().list,
            result_key='users',
            params=dict(
                customer='my_customer',  # TODO: double check this param
                domain=domain,
                orderBy='email',
            )
        )

    def get_groups(self, domain: str = None) -> Iterator[GroupType]:
        yield from self._paginate(
            method=self.client.groups().list,
            result_key='groups',
            params=dict(
                domain=domain,
            )
        )

    def get_group(self, group_id: str) -> GroupType:
        return self.client.groups().get(
            groupKey=group_id,
        ).execute()

    def create_or_update_group(
            self,
            email: str,
            name: str = None,
            description: str = None,
            group_id: str = None,
    ) -> GroupType:
        body = dict(
            email=email,
            name=name,
            description=description,
        )

        if not group_id:
            return self.client.groups().insert(
                body=body,
            ).execute()
        else:
            return self.client.groups().update(
                groupKey=group_id,
                body=body,
            ).execute()

    def delete_group(self, group_id: str) -> GroupType:
        try:
            return self.client.groups().delete(
                groupKey=group_id,
            ).execute()
        except HttpError as e:
            if e.resp.status == 404:
                raise exceptions.NotFoundError()
            raise e

    def get_group_members(self, group_id: str) -> Iterator[MemberType]:
        page_token = None

        while True:
            results = self.client.members().list(
                groupKey=group_id,
                pageToken=page_token,
            ).execute()
            for member in results.get('members', []):
                yield member

            page_token = results.get('nextPageToken')
            if not page_token:
                break

    def add_group_member(self, group_id: str, email: str, role: str = 'MEMBER') -> MemberType:
        body = {'email': email, 'role': role}

        try:
            return self.client.members().insert(
                groupKey=group_id,
                body=body,
            ).execute()
        except HttpError as e:
            if e.resp.status == 409:
                raise exceptions.AlreadyExistsError()

            if e.resp.status == 404:
                raise exceptions.NotFoundError()

            raise e

    def delete_group_member(self, group_id: str, member_id: str) -> MemberType:
        try:
            return self.client.members().delete(
                groupKey=group_id,
                memberKey=member_id,
            ).execute()
        except HttpError as e:
            if e.resp.status == 404:
                raise exceptions.NotFoundError()
            raise e
